#include <vector>
#include <map>
#include <iostream>
#include <sstream>
#include <ctime>

#include <signal.h>

#include <windows.h>

#include "common/library.h"
#include "common/thread.h"

#include "platform/vfw.h"

#include "avisynth/compatible.h"

#include "aviutl/filter.h"
#include "aviutl/output.h"

namespace CURRENT_VERSION {

#include "aviutl.h"

#define IMPLEMENT
#include "loadauo.h"
#undef IMPLEMENT

class OutputCallback : public AviUtlOutputCallback {
	const char *_auo, *_cfg, *_avs, *_out;
	int _now, _total;
	bool _abort, _result;

	IScriptEnvironment *_env;
	time_t _start, _finish;

	AviUtlOutputPluginRef _plugin;
	avisynth::AvisynthDLLRef _avisynth;
	avisynth::EnvironmentObjectRef _holder;
	PClip _clip, _yuy2, _rgb24;
	PVideoFrame _videoBuf;
	std::vector<BYTE> _audioBuf;

	static const char *GetImporter(const char *file) {
		using namespace std;

		const char *importer = "Import";
		DWORD header[3] = { 0, 0, 0 };

		ifstream ifs(file, ios_base::binary);
		ifs.read(reinterpret_cast<char *>(header), sizeof(header));

		if(header[0] == mmioFOURCC('R','I','F','F') && header[2] == mmioFOURCC('A','V','I',' '))
			importer = "OpenDMLSource";

		return importer;
	}

protected:
	bool Init() {
		using namespace std;

		_plugin = new AviUtlOutputPlugin(_auo, "", false);

		if(!_plugin) {
			cout << "vOCǍG[" << endl;
			return false;
		}

		if(!_plugin->Init()) {
			cout << "vOCG[" << endl;
			return false;
		}

		if(!_plugin->Config(_cfg)) {
			cout << "vOCݒG[" << endl;
			return false;
		}

		_avisynth = new avisynth::AvisynthDLL("avisynth.dll");

		if(!_avisynth) {
			cout << "avisynth.dllǍ߂Ȃ" << endl;
			return false;
		}

		_holder = _avisynth->CreateScriptEnvironment();

		if(!_holder) {
			cout << "ScriptEnvironment\zoȂ" << endl;
			return false;
		}

		_env = _holder->CreateCompatible();

		if(_env == NULL) {
			cout << "ScriptEnvironmenť݊ۂĂȂ" << endl;
			return false;
		}

		const char *importer = GetImporter(_avs);

		try {
			AVSValue args[] = { _avs };
			_clip = _env->Invoke(importer, AVSValue(args, sizeof(args) / sizeof(*args))).AsClip();
		}
		catch(...) {
			cout << "Nbv\zɗO" << endl;
			return false;
		}

		if(!_clip) {
			cout << "Nbv擾oȂ" << endl;
			return false;
		}

		try {
			AVSValue args[] = { _clip };
			_yuy2 = _env->Invoke("ConvertToYUY2", AVSValue(args, sizeof(args) / sizeof(*args))).AsClip();
			_rgb24 = _env->Invoke("ConvertToRGB24", AVSValue(args, sizeof(args) / sizeof(*args))).AsClip();
		}
		catch(...) {
			cout << "Ro[^\zɗO" << endl;
			return false;
		}

		return true;
	}

	void Exit() {
		_audioBuf.clear();
		_videoBuf = PVideoFrame();
		_rgb24 = PClip();
		_yuy2 = PClip();
		_clip = PClip();
		_holder = avisynth::EnvironmentObjectRef();
		_avisynth = avisynth::AvisynthDLLRef();
		_plugin = AviUtlOutputPluginRef();
	}

	void Start() {
		using namespace std;

		cout << "vOC: " << _auo << endl;
		cout << "ݒt@C: " << _cfg << endl;
		cout << "̓t@C: " << _avs << endl;
		cout << "o̓t@C: " << _out << endl;
		cout << endl;

		time(&_start);
		cout << "GR[hJn: " << ctime(&_start) << endl;
	}

	void Finish(bool result) {
		using namespace std;

		time(&_finish);
		cout << "GR[hI: " << ctime(&_finish) << endl;

		int hour = difftime(_finish, _start);
		int second = hour % 60; hour /= 60;
		int minute = hour % 60; hour /= 60;

		cout << "oߎ: " << hour << ":" << minute << ":" << second << endl;

		_result = result;
	}

	void GetOutputInfo(OUTPUT_INFO *oip) {
		const VideoInfo& vi = _rgb24->GetVideoInfo();
		oip->savefile = const_cast<char *>(_out);

		if(vi.HasVideo()) {
			oip->flag |= OUTPUT_INFO_FLAG_VIDEO;
			oip->w = vi.width;
			oip->h = vi.height;
			oip->rate = vi.fps_numerator;
			oip->scale = vi.fps_denominator;
			oip->n = vi.num_frames;
			oip->size = vi.BMPSize();
		}

		if(vi.HasAudio()) {
			oip->flag |= OUTPUT_INFO_FLAG_AUDIO;
			oip->audio_rate = vi.SamplesPerSecond();
			oip->audio_ch = vi.AudioChannels();
			oip->audio_n = vi.num_audio_samples;
			oip->audio_size = vi.BytesPerAudioSample();
		}
	}

	const AviUtlOutputPluginRef& GetPlugin() {
		return _plugin;
	}

	void *GetVideo(int frame, DWORD format) {
		PClip clip;

		switch(format) {
		case mmioFOURCC('Y','U','Y','2'):
			clip = _yuy2;
			break;

		case BI_RGB: case comptypeDIB:
			clip = _rgb24;
			break;

		default:
			return NULL;
		}

		const VideoInfo& vi = clip->GetVideoInfo();
		_videoBuf = clip->GetFrame(frame, _env);

		if(_videoBuf->GetPitch() != ((vi.RowSize() + 3) & ~3)) {
			PVideoFrame dst(_env->NewVideoFrame(vi, 4));

			_env->BitBlt(dst->GetWritePtr(), dst->GetPitch(),
				_videoBuf->GetReadPtr(), _videoBuf->GetPitch(),
				_videoBuf->GetRowSize(), _videoBuf->GetHeight());

			_videoBuf = dst;
		}

		return const_cast<BYTE *>(_videoBuf->GetReadPtr());
	}

	void *GetAudio(int start, int length, int *readed) {
		const VideoInfo& vi = _clip->GetVideoInfo();

		_audioBuf.resize(vi.BytesFromAudioSamples(length));
		_clip->GetAudio(&_audioBuf[0], start, length, _env);

		if(readed != NULL)
			*readed = length;

		return &_audioBuf[0];
	}

	bool IsAbort() {
		return _abort;
	}

	bool RestTimeDisp(int now, int total) {
		_now = now;
		_total = total;
		return true;
	}

	int GetFlag(int frame) {
		return 0;
	}

	bool UpdatePreview() {
		return true;
	}

public:
	OutputCallback(const char *auo, const char *cfg, const char *avs, const char *out)
		: _auo(auo), _cfg(cfg), _avs(avs), _out(out)
		, _now(0), _total(std::numeric_limits<int>::max())
		, _abort(false), _result(false) {}

	void Abort()
	{ _abort = true; }

	template<class T>
	void GetProgress(T& now, T& total) const
	{ now = _now; total = _total; }

	bool GetResult() const
	{ return _result; }
};

} // namespace CURRENT_VERSION

//////////////////////////////////////////////////////////////////////////////

namespace CURRENT_VERSION {

std::auto_ptr<CURRENT_VERSION::OutputCallback> _callback;

static void OnSignal(int signal) {
	switch(signal) {
	case SIGINT:
		_callback->Abort();
		break;
	}
}

} // namespace CURRENT_VERSION

//////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv) {
	using namespace CURRENT_VERSION;
	using namespace std;

	static const DWORD priority[] = {
		IDLE_PRIORITY_CLASS, BELOW_NORMAL_PRIORITY_CLASS,
		NORMAL_PRIORITY_CLASS, ABOVE_NORMAL_PRIORITY_CLASS,
		HIGH_PRIORITY_CLASS, REALTIME_PRIORITY_CLASS,
	};
	const int prioritySize = sizeof(priority) / sizeof(*priority);

	if(argc != 6) {
		const char *command = strrchr(argv[0], '\\');

		if(command == NULL)
			command = argv[0];

		cout << "Usage: " << command;
		cout << " AUO̧ ݒ̧ ̧ o̧ ײè(0-";
		cout << (prioritySize - 1) << ")" << endl;

		return EXIT_FAILURE;
	}

	if(isdigit(argv[5][0]) && argv[5][1] == '\0') {
		int index = (argv[5][0] - '0') % prioritySize;
		SetPriorityClass(GetCurrentProcess(), priority[index]);
	}

	_callback.reset(new OutputCallback(argv[1], argv[2], argv[3], argv[4]));
	signal(SIGINT, OnSignal);

	AviUtlOutputContextRef context(new AviUtlOutputContext(_callback.get()));
	context->Start();

	while(!context->Wait(1000)) {
		int now, total;
		_callback->GetProgress(now, total);

		ostringstream os;
		os << now << "/" << total;

		if(total != 0)
			os << " (" << (now * 100 / total) << "%)";

		SetConsoleTitle(os.str().c_str());
	}

	if(!_callback->GetResult()) {
		cout << "GR[h͎s܂" << endl;
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}

//////////////////////////////////////////////////////////////////////////////
